/*
*	SmallBASIC platform driver for Unix|Win32 + SDL
*
*	ndc: 2001-03-03
*/

#include "device.h"
#include "osd.h"
#include "str.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(_UnixOS)
#include <unistd.h>
#endif
#include <SDL/SDL.h>

static int	cur_x = 0;
static int	cur_y = 0;
static int	bytespp = 1;
static int	mouse_mode, mouse_x, mouse_y, mouse_b, mouse_upd, mouse_down_x, mouse_down_y, mouse_pc_x, mouse_pc_y;
static int	tabsize = 32;	// from dev_palm
static int	maxline;
static int	font_h = 16;

static int	con_use_bold = 0;
static int	con_use_ul   = 0;
static int	con_use_reverse = 0;

// VGA16 colors in RGB
static unsigned long vga16[] =
{
0x0, 0x7F, 0x7F00, 0x7F7F, 0x7F0000, 0x7F007F, 0x7F7F00, 0x808080,
0x555555, 0xFF, 0xFF00, 0xFFFF, 0xFF0000, 0xFF00FF, 0xFFFF00, 0xFFFFFF 
};

static int	cmap[16];

// fonts
#include "unix/rom16.c"

static SDL_Surface	*screen;
static int has_audio = 0;
static int mouse_hot_x = -16, mouse_hot_y = -16;
static SDL_AudioSpec audiospec;

//#define	TAKE_SCREENSHOT

#if defined(TAKE_SCREENSHOT)
void	writeppm(char *name);
#endif

struct voice_info {
	float	period;
	float	toggle;
	int		setting;
	int		vol;
	};

static char* mixbuf;
#define	AUDIO_STACK_MAX		8192*2
static struct voice_info	audio_info[AUDIO_STACK_MAX];
static int audio_head = 0;
static int audio_tail = 0;

/*
*	I know nothing about sound.
*	someone to fix it
*/
void	audio_callback(void *user, unsigned char* stream, int length)
{
	int		volume;
	int		mix = 0;
	struct voice_info *info;

    memset(mixbuf, audiospec.silence, length);
	if ( audio_head != audio_tail )	{
		info = &audio_info[audio_head];
		volume = (info->vol * SDL_MIX_MAXVOLUME) / 100;

        if ( info->period >= 1.0 ) {
            int		left = length, j = 0;
			int		count;

	        mix = 1;
            do {
				count = (info->toggle < left) ? (int) info->toggle : left;
                left -= count;
                info->toggle -= count;
                while ( count -- ) 
                    mixbuf[j++] += info->setting;

                if ( info->toggle < 1.0 ) {
                    info->toggle += info->period;
					info->setting = -info->setting;
                    }
	            } while ( left > 0 );

			}			  

		audio_head ++;
		if	( audio_head >= AUDIO_STACK_MAX )
			audio_head = 0;

	    if ( mix )
			SDL_MixAudio(stream, mixbuf, length, volume);
	    }
}

/*
*	SDL: Initialization
*/
int		osd_devinit()
{
	SDL_Color colors[256];
	int		i;

	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "SDL: Couldn't initialize SDL: %s\n", SDL_GetError());
		return 0;
	    }
	if ( ( screen = SDL_SetVideoMode(640, 480, 16, 0) ) == NULL ) {
		fprintf(stderr, "SDL: Couldn't set 640x480 video mode: %s\n",  SDL_GetError());
		return 0;
	    }
	has_audio = (SDL_InitSubSystem(SDL_INIT_AUDIO) >= 0);

	os_graf_mx = screen->w;
	os_graf_my = screen->h;
	maxline = os_graf_my / font_h;
	os_graphics = DEVDRV_SDL;
	os_color_depth = screen->format->BitsPerPixel;
	os_color = 1;
	if	( os_color_depth != 8 && os_color_depth != 16 && os_color_depth != 32 )	{
		fprintf(stderr, "SDL: Couldn't set 640x480 video mode\n");
		return 0;
	    }

	SDL_WM_SetCaption("SmallBASIC/SDL", NULL);
		
	bytespp = screen->format->BytesPerPixel;

	if	( os_color_depth == 8 )	{
		for ( i = 0; i < 256; i ++)	{
			colors[i].r=i;
			colors[i].g=i;
			colors[i].b=i;
			}
		for ( i = 0; i < 16; i ++)	{
			colors[i].r=(vga16[i] & 0xFF0000) >> 16;
			colors[i].g=(vga16[i] & 0xFF00) >> 8;
			colors[i].b=(vga16[i] & 0xFF);
			cmap[i] = i;
			}
		SDL_SetColors(screen, colors, 0, 256);
		}
	else	{
		for ( i = 0; i < 16; i ++ )	{
			cmap[i] = SDL_MapRGB(screen->format, 
				(vga16[i] & 0xFF0000) >> 16,
				(vga16[i] & 0xFF00) >> 8,
				(vga16[i] & 0xFF) );
			}
		}

	if	( has_audio )	{
		SDL_AudioSpec	wanted;

		memset(&wanted, 0, sizeof(SDL_AudioSpec));

		// Set the audio format
		wanted.freq     = 44100;
		wanted.format   = AUDIO_U8;
		wanted.channels = 1;
		wanted.samples  = 4096;
		wanted.callback = audio_callback;
		wanted.userdata = NULL;

		// Open the audio device, forcing the desired format
		if ( SDL_OpenAudio ( &wanted, &audiospec ) < 0 ) 
			has_audio = 0;
		else	{
			mixbuf = (char *) malloc(0x10000);
			SDL_PauseAudio(0);
			}
		}

	#if defined(_UnixOS)
	setsysvar_str(SYSVAR_OSNAME, "Unix/SDL");
	#else
	setsysvar_str(SYSVAR_OSNAME, "Win32/SDL");
	#endif
	osd_cls();

	return 1;
}

/*
*	close
*/
int		osd_devrestore()
{
	cur_x = 0;
	cur_y = (maxline - 1) * font_h;
	osd_write("SDL: Press any key to exit...");
#if defined(TAKE_SCREENSHOT)
	writeppm("sdl.ppm");	
#endif
	SDL_PauseAudio(1);
	while ( !dev_kbhit() );
	SDL_Quit();
	if	( mixbuf )
		free(mixbuf);

	return 1;
}


//
void	sdl_setpixel(int x, int y, int c)
{
	int		offset;

	if	( x >= 0 && x < screen->w )	{
		if	( y >= 0 && y < screen->h )		{
			offset = y * screen->w + x;
			switch ( bytespp )	{
			case	1:
				*(((byte *)screen->pixels)+offset) = c;
				break;
			case	2:
				*(((short int *)screen->pixels)+offset) = c;
				break;
			case	4:
				*(((long int *)screen->pixels)+offset) = c;
				break;
				}
			}
		}
}

//
//#define SDL_FASTPIX(x,y)	*(((byte *)screen->pixels)+(y)*screen->w+(x)) = dev_fgcolor
#define SDL_FASTPIX(x,y)	sdl_setpixel(x,y,cmap[dev_fgcolor])

/* Bresenham's algorithm for drawing line */
void	sdl_line(int X1, int Y1, int X2, int Y2)
{
	int		dX, dY, row, col, final, G, inc1, inc2;
	char	pos_slope;

	dX = X2 - X1;
	dY = Y2 - Y1;
	pos_slope = (dX > 0);
	if ( dY < 0 )
		pos_slope = !pos_slope;

	if ( ABS(dX) > ABS(dY) )	{
		if (dX > 0)    {
			col = X1;
			row = Y1;
			final = X2;
		    }
		else	{
			col = X2;
			row = Y2;
			final = X1;
		    }

	    inc1 = 2 * abs (dY);
	    G = inc1 - abs (dX);
	    inc2 = 2 * (abs (dY) - abs (dX));
	    if ( pos_slope )	{
			while (col <= final)	{
				SDL_FASTPIX(col, row);
				col++;
				if (G >= 0) {
					row ++;
					G += inc2;
					}
				else		
					G += inc1;
				}
			}
		else	{
			while (col <= final)	{
				SDL_FASTPIX(col, row);
				col ++;
				if (G > 0)	{
					row --;
					G += inc2;
					}
				else
					G += inc1;
				}
			}
		}  /* if |dX| > |dY| */
	else	{
		if (dY > 0)	{
			col = X1;
			row = Y1;
			final = Y2;
		    }
		else	{
			col = X2;
			row = Y2;
			final = Y1;
		    }

		inc1 = 2 * abs (dX);
		G = inc1 - abs (dY);
		inc2 = 2 * (abs (dX) - abs (dY));

		if (pos_slope)	{
			while (row <= final) {
				SDL_FASTPIX(col, row);
				row ++;
				if (G >= 0)	{
					col ++;
					G += inc2;
					}
				else
					G += inc1;
				}
			}
		else	{
			while (row <= final)	{
				SDL_FASTPIX(col, row);
				row ++;
				if (G > 0)	{
					col --;
	  				G += inc2;
					}
				else
					G += inc1;
		     	}
			}
		}
}

void	sdl_fillrect(int x1, int y1, int x2, int y2, int c)
{
	int		i, j, dx, count;
	int		offset;

	if	( y2 > screen->h )
		y2 = screen->h;
	if	( x2 > screen->w )
		x2 = screen->w;

	dx = x2 - x1;
	count = dx * bytespp;
	for ( i = y1; i < y2; i ++ )	{
		offset = i * screen->w * bytespp + x1 * bytespp;
		switch ( bytespp )	{
		case	1:
			memset((char *)screen->pixels+offset, c, count);
			break;
		case	2:
			for ( j = 0; j < dx; j ++ )	{
				memcpy((char *)screen->pixels+offset, &c, 2);
				offset += 2;
				}
			break;
		case	4:
			for ( j = 0; j < dx; j ++ )	{
				memcpy((char *)screen->pixels+offset, &c, 4);
				offset += 4;
				}
			break;
			}
		}
}

void	osd_settextcolor(int fg, int bg)
{
	osd_setcolor(fg);
	if	( bg != -1 )
		dev_bgcolor = bg;
}

//
void	osd_drawchar(int x, int y, byte ch, int overwrite, int fg_rgb, int bg_rgb)
{
	int		offset;
	int		bit, i;

	offset = ch * 16;

	for ( i = 0; i < 16; i ++, offset ++ )	{
		for ( bit = 0; bit < 8; bit ++ )	{
			if	( *(font8x16+offset) & (1 << (8-bit)) )
				sdl_setpixel(x+bit, y+i, fg_rgb);
			else if ( overwrite )
				sdl_setpixel(x+bit, y+i, bg_rgb);
			}
		}
}

/*
*	enable or disable PEN code
*/
void	osd_setpenmode(int enable)
{
 	mouse_mode = enable;
}

/*
*/
int		osd_getpen(int code)
{
	int		r = 0;
	byte	update;

	osd_events(0);	
	if	( mouse_mode )	{
		switch ( code )	{
		case 	0:	// bool: status changed
			r = mouse_upd;
			break;		
		case	1:	// last pen-down x
			r = mouse_down_x;
			break;		
		case	2:	// last pen-down y
			r = mouse_down_y;
			break;		
		case	3:	// vert. 1 = down, 0 = up .... unsupported
			r = 0;
			break;
		case	4:	// last x
			r = mouse_pc_x;
			break;
		case	5:	// last y
			r = mouse_pc_y;
			break;
		case	10:
			r = mouse_x;
			break;
		case	11:
			r = mouse_y;
			break;
		case	12:
		case	13:
		case	14:
			r = (mouse_b & (1 << (code - 11))) ? 1 : 0;
			break;
			}

		mouse_upd = 0;
		}
	return	r;
}

//
void	osd_cls()
{
	cur_x = cur_y = 0;

	SDL_LockSurface(screen);
	sdl_fillrect(0,0,screen->w,screen->h,cmap[15]);
	SDL_UnlockSurface(screen);
//	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

//	returns the current x position
int		osd_getx()
{
	return cur_x;
}

//	returns the current y position
int		osd_gety()
{
	return cur_y;
}

//
void	osd_setxy(int x, int y)
{
	cur_x = x;
	cur_y = y;
}

/**
*	next line
*/
void	osd_nextln()
{
	cur_x = 0;

	if	( cur_y < ((maxline-1) * font_h) )	{
		cur_y += font_h;
		}
	else	{
		char	*buf;
		int		len;

		// scroll
//		gl_copybox(0, font_h, os_graf_mx, cur_y, 0, 0);
		len = cur_y * bytespp * os_graf_mx;
		buf = (char *) malloc(len);
		memcpy(buf, (char*) screen->pixels + font_h*bytespp*os_graf_mx, len);
		memcpy(screen->pixels, buf, len);
		free(buf);

		sdl_fillrect(0, cur_y, os_graf_mx, font_h, cmap[15]);
		}
}

/*
*	calc next tab position
*/
int		osd_calctab(int x)
{
	int		c = 1;

	while ( x > tabsize )	{
		x -= tabsize;
		c ++;
		}
	return c * tabsize;
}

/**
*	Basic output
*
*	Supported control codes:
*	\t		tab (20 px)
*	\a		beep
*	\n		next line (cr/lf)
*	\xC		clear screen
*	\e[K	clear to end of line
*	\e[0m	reset all attributes to their defaults
*	\e[1m	set bold on
*	\e[4m	set underline on
*	\e[7m	reverse video
*	\e[21m	set bold off
*	\e[24m	set underline off
*	\e[27m	set reverse off
*/
void	osd_write(const char *str)
{
	int		len, cx = 8, esc_val, esc_cmd;
	byte	*p, buf[3];
	int		ch, char_len = 1;

	len = strlen(str);

	if	( len <= 0 )
		return;

	SDL_LockSurface(screen);

	p = (byte *) str;
	while ( *p )	{
		switch ( *p )	{
		case	'\a':	// beep
			osd_beep();
			break;
		case	'\t':
			cur_x = osd_calctab(cur_x+1);
			break;
		case	'\xC':
			osd_cls();
			break;
		case	'\033':		// ESC ctrl chars (man console_codes)
			if	( *(p+1) == '[' )	{
				p += 2;
				esc_val = esc_cmd = 0;

				if	( is_digit(*p) )	{
					esc_val = (*p - '0');
					p ++;

					if	( is_digit(*p) )	{
						esc_val = (esc_val * 10) + (*p - '0');
						p ++;
						}

					esc_cmd = *p;
					}
				else	
					esc_cmd = *p;

				// control characters
				switch ( esc_cmd )	{
				case	'K':			// \e[K - clear to eol
					sdl_fillrect(cur_x, cur_y, os_graf_mx - cur_x, cur_y+font_h, cmap[15]);
					break;
				case	'm':			// \e[...m	- ANSI terminal
					switch ( esc_val )	{
					case	0:	// reset
						con_use_bold = 0;
						con_use_ul = 0;
						con_use_reverse = 0;
						osd_setcolor(0);
						osd_settextcolor(0, 15);
						break;
					case	1:	// set bold on
						con_use_bold = 1;
						break;
					case	4:	// set underline on
						con_use_ul = 1;
						break;
					case	7:	// reverse video on
						con_use_reverse = 1;
						break;
					case	21:	// set bold off
						con_use_bold = 0;
						break;
					case	24:	// set underline off
						con_use_ul = 0;
						break;
					case	27:	// reverse video off
						con_use_reverse = 0;
						break;

					// colors - 30..37 foreground, 40..47 background
					case	30:	// set black fg
						osd_setcolor(0);
						break;
					case	31:	// set red fg
						osd_setcolor(4);
						break;
					case	32:	// set green fg
						osd_setcolor(2);
						break;
					case	33:	// set brown fg
						osd_setcolor(6);
						break;
					case	34:	// set blue fg
						osd_setcolor(1);
						break;
					case	35:	// set magenta fg
						osd_setcolor(5);
						break;
					case	36:	// set cyan fg
						osd_setcolor(3);
						break;
					case	37:	// set white fg
						osd_setcolor(7);
						break;

					case	40:	// set black bg
						osd_settextcolor(cmap[dev_fgcolor], 0);
						break;
					case	41:	// set red bg
						osd_settextcolor(cmap[dev_fgcolor], 4);
						break;
					case	42:	// set green bg
						osd_settextcolor(cmap[dev_fgcolor], 2);
						break;
					case	43:	// set brown bg
						osd_settextcolor(cmap[dev_fgcolor], 6);
						break;
					case	44:	// set blue bg
						osd_settextcolor(cmap[dev_fgcolor], 1);
						break;
					case	45:	// set magenta bg
						osd_settextcolor(cmap[dev_fgcolor], 5);
						break;
					case	46:	// set cyan bg
						osd_settextcolor(cmap[dev_fgcolor], 3);
						break;
					case	47:	// set white bg
						osd_settextcolor(cmap[dev_fgcolor], 7);
						break;

						};
					break;
					}
				}
			break;
		case	'\n':		// new line
			SDL_UnlockSurface(screen);
			osd_nextln();
			SDL_LockSurface(screen);
			break;
		case	'\r':		// return
			cur_x = 0;
			sdl_fillrect(cur_x, cur_y, os_graf_mx - cur_x, cur_y+font_h, cmap[15]);
			break;
		default:
			//
			//	PRINT THE CHARACTER
			//
			buf[0] = *p;
			buf[1] = '\0';

			// new line ?
			if	( cur_x + cx >= dev_x2clip )
				osd_nextln();

			// draw

			// TODO: ??? SJIS on Linux ???
			if	( !con_use_reverse )	{
				osd_drawchar(cur_x, cur_y, *p, 1, cmap[dev_fgcolor], cmap[dev_bgcolor]);
				if	( con_use_bold )	
					osd_drawchar(cur_x-1, cur_y, *p, 0, cmap[dev_fgcolor], cmap[dev_bgcolor]);
				}
			else	{
				osd_drawchar(cur_x, cur_y, *p, 1, cmap[dev_bgcolor], cmap[dev_fgcolor]);
				if	( con_use_bold )	
					osd_drawchar(cur_x-1, cur_y, *p, 0, cmap[dev_bgcolor], cmap[dev_fgcolor]);
				}

			if	( con_use_ul )	{
				osd_setcolor(dev_fgcolor);
				sdl_line(cur_x, (cur_y+font_h)-1, cur_x+cx, (cur_y+font_h)-1);
				}

			// advance
			cur_x += cx;
			};

		if	( *p == '\0' )
			break;

		p ++;
		}

	SDL_UnlockSurface(screen);
//	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

int		osd_events(int wait_flag)
{
	int			ch, button, upr;
	int			evc = 0;
	SDL_Event	ev;

	SDL_UpdateRect(screen, 0, 0, 0, 0);

	if	( wait_flag == 0 )	{
		if	( SDL_PollEvent(&ev) )	{
			switch ( ev.type )	{
			case	SDL_KEYDOWN:
				ch = ev.key.keysym.sym;
				if	( ch == SDLK_c && (ev.key.keysym.mod & KMOD_CTRL) )	{
					return -2;
					}
				else if	( ch == SDLK_CAPSLOCK )	
					;
				else if	( ch == SDLK_LSHIFT || ch == SDLK_RSHIFT  )		
					;
				else if	( ch == SDLK_LALT || ch == SDLK_RALT  )		
					;
				else if	( ch == SDLK_LCTRL || ch == SDLK_RCTRL  )		
					;
				else if	( ch == SDLK_RETURN )
					dev_pushkey('\n');
				else if	( ch == SDLK_BACKSPACE )
					dev_pushkey(8);
				else if	( ch == 127 )
					dev_pushkey(8);
				else	{
					upr = ( ((ev.key.keysym.mod & (KMOD_CAPS))) && ((ev.key.keysym.mod & (KMOD_SHIFT)) == 0) )
							||
						  ( ((ev.key.keysym.mod & (KMOD_CAPS)) == 0) && ((ev.key.keysym.mod & KMOD_SHIFT)) );

					if	( upr )
						dev_pushkey(toupper(ch));
					else
						dev_pushkey(ch);
					}
				evc ++;
				break;
			case	SDL_MOUSEBUTTONDOWN:
			case	SDL_MOUSEBUTTONUP:
			case	SDL_MOUSEMOTION:
				button = SDL_GetMouseState(&mouse_x, &mouse_y);

				if	( mouse_x < mouse_hot_x )		mouse_x = mouse_hot_x;
				if	( mouse_y < mouse_hot_y )		mouse_y = mouse_hot_y;
				if	( mouse_x >= os_graf_mx - 1 )	mouse_x = os_graf_mx - 1;
				if	( mouse_y >= os_graf_my - 1 )	mouse_y = os_graf_my - 1;
//				vga_setcursorposition(mouse_x + mouse_hot_x, mouse_y + mouse_hot_y);
				mouse_b = 0;	/// bug
				if ( button & SDL_BUTTON(1) )	{
					if	( (mouse_b & SDL_BUTTON(1)) == 0 )	{	// new press
						mouse_down_x = mouse_x;
						mouse_down_y = mouse_y;
						}

					mouse_upd = 1;

					mouse_pc_x = mouse_x;
					mouse_pc_y = mouse_y;

					mouse_b |= 1;
					}
				if (button & SDL_BUTTON(2))
					mouse_b |= 2;
				if (button & SDL_BUTTON(3))
					mouse_b |= 4;

				evc ++;
				}
			}
		}
	else	{
		while ( (evc = osd_events(0)) == 0 )
			SDL_Delay(50);
		}

	return evc;
}

///////////////////////////////////////////////////////////////

void	osd_setcolor(int color)
{
	dev_fgcolor = color;
}

void	osd_line(int x1, int y1, int x2, int y2)
{
	SDL_LockSurface(screen);

	if	( (x1 == x2) && (y1 == y2) )
		sdl_setpixel(x1, y1, cmap[dev_fgcolor]);
	else
		sdl_line(x1, y1, x2, y2);

	SDL_UnlockSurface(screen);
//	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

void	osd_setpixel(int x, int y)
{
	SDL_LockSurface(screen);

	sdl_setpixel(x, y, cmap[dev_fgcolor]);

	SDL_UnlockSurface(screen);
//	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

void	osd_rect(int x1, int y1, int x2, int y2, int fill)
{
	int		y;

	SDL_LockSurface(screen);

	if	( fill )	{
		for ( y = y1; y <= y2; y ++ )	
			sdl_line(x1, y, x2, y);
		}
	else	{
		sdl_line(x1, y1, x2, y2);
		sdl_line(x1, y1, x1, y2);
		sdl_line(x1, y2, x2, y2);
		sdl_line(x2, y2, x2, y1);
		sdl_line(x2, y1, x1, y1);
		}

	SDL_UnlockSurface(screen);
//	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

///////////////////////////////////////////////////////////////

void	osd_sound(int freq, int ms, int vol, int bgplay)
{
	struct voice_info *info;
	int		i, loops;

	if	( has_audio )	{

		if	( !freq )
			freq = audiospec.freq / 2;

		loops = ((ms * audiospec.freq / freq / 2.0) / audiospec.samples) / 1000;

		SDL_LockAudio();

		info = &audio_info[audio_tail];
		info->vol = vol;
		info->setting = 0;
		info->period = 0;
	  	info->toggle = 1;

        if ( (freq < audiospec.freq / 2) && (freq != 0) ) {
            info->period  = (float) (audiospec.freq / freq / 2.0);
            info->setting = (info->setting > 0) ? info->vol : -info->vol;
	        } 
		else 	
            info->period  = 0.0;

		// 
		audio_tail ++;
		if	( audio_tail >= AUDIO_STACK_MAX )
			audio_tail = 0;

		for ( i = 1; i < loops; i ++ )	{
			audio_info[audio_tail] = *info;
			audio_info[audio_tail].toggle = 0;

			audio_tail ++;
			if	( audio_tail >= AUDIO_STACK_MAX )
				audio_tail = 0;
			}

		SDL_UnlockAudio();

		if	( !bgplay )
			SDL_Delay(ms);
		}
}

void	osd_beep()
{
	if	( has_audio )
		osd_sound(440, 250, 75, 0);
	else
		printf("\a");
}

///////////////////////////////////////////////////////////////

int		osd_textwidth(const char *str)
{
	int		l = strlen(str);

	// SJIS ???
	return l * 8;
}

int		osd_textheight(const char *str)
{
	// TODO: count \n
	return font_h;
}

#if defined(TAKE_SCREENSHOT)
//
void	writeppm(char *name)
{
    int		y, i, j, k;
	FILE	*file;
	byte	*tmp, *tmp2;
	int		width = os_graf_mx;
	int		height = os_graf_my;
	int		bpp = os_color_depth;

	file = fopen(name, "wt");
    fprintf(file, "P6\n%i %i\n255\n", width, height);

	tmp = (byte *) malloc(width*4);
	tmp2 = (byte *) malloc(width*4);

    for ( y = 0; y < height; y++ )	{

        memcpy(tmp, screen->pixels+(screen->w*y*bytespp), width * bytespp);

        switch ( bpp ) {
        case 16:
            for ( i = 0; i < width; i++ ) {
                j = tmp[i * 2] + 256 * tmp[i * 2 + 1];
                tmp2[i * 3] = (j & 0xf800) >> 8;
                tmp2[i * 3 + 1] = (j & 0x7e0) >> 3;
                tmp2[i * 3 + 2] = (j & 0x1f) << 3;
	            }
            break;
        case 24:
        case 32:
            memcpy(tmp2, tmp, width * 3);
            break;
	        }

        fwrite(tmp2, width * 3, 1, file);
		}

	free(tmp);
	free(tmp2);

	fclose(file);
}
#endif
